@@ -1,10 +1,11 @@
 /*
  * Pextlib.c
- * $Id: Pextlib.c 82292 2011-08-12 09:05:23Z afb@macports.org $
+ * $Id: Pextlib.c 89818 2012-02-11 10:13:16Z and.damore@macports.org $
  *
  * Copyright (c) 2002 - 2003 Apple Inc.
  * Copyright (c) 2004 - 2005 Paul Guyot <pguyot@kallisys.net>
  * Copyright (c) 2004 Landon Fuller <landonf@macports.org>
+ * Copyright (c) 2012 The MacPorts Project
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -46,6 +47,7 @@
 #include <grp.h>
 #include <limits.h>
 #include <pwd.h>
+#include <stdbool.h>
 #include <stdint.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -53,6 +55,11 @@
 #include <strings.h>
 #include <unistd.h>
 
+#ifdef __MACH__
+#include <mach-o/loader.h>
+#include <mach-o/fat.h>
+#endif
+
 #include <tcl.h>
 
 #include "Pextlib.h"
@@ -93,147 +100,147 @@ extern char **environ;
 static char *
 ui_escape(const char *source)
 {
-	char *d, *dest;
-	const char *s;
-	size_t dlen;
+    char *d, *dest;
+    const char *s;
+    size_t dlen;
 
-	s = source;
-	dlen = strlen(source) * 2 + 1;
-	d = dest = malloc(dlen);
-	if (dest == NULL) {
-		return NULL;
-	}
-	while(*s != '\0') {
-		switch(*s) {
-			case '\\':
-			case '}':
-			case '{':
-				*d = '\\';
-				d++;
-				*d = *s;
-				d++;
-				s++;
-				break;
-			case '\n':
-				s++;
-				break;
-			default:
-				*d = *s;
-				d++;
-				s++;
-				break;
-		}
-	}
-	*d = '\0';
-	return dest;
+    s = source;
+    dlen = strlen(source) * 2 + 1;
+    d = dest = malloc(dlen);
+    if (dest == NULL) {
+        return NULL;
+    }
+    while(*s != '\0') {
+        switch(*s) {
+            case '\\':
+            case '}':
+            case '{':
+                *d = '\\';
+                d++;
+                *d = *s;
+                d++;
+                s++;
+                break;
+            case '\n':
+                s++;
+                break;
+            default:
+                *d = *s;
+                d++;
+                s++;
+                break;
+        }
+    }
+    *d = '\0';
+    return dest;
 }
 
 int
 ui_info(Tcl_Interp *interp, char *mesg)
 {
-	const char ui_proc_start[] = "ui_info [subst -nocommands -novariables {";
-	const char ui_proc_end[] = "}]";
-	char *script, *string;
-	size_t scriptlen, len, remaining;
-	int rval;
+    const char ui_proc_start[] = "ui_info [subst -nocommands -novariables {";
+    const char ui_proc_end[] = "}]";
+    char *script, *string;
+    size_t scriptlen, len, remaining;
+    int rval;
 
-	string = ui_escape(mesg);
-	if (string == NULL)
-		return TCL_ERROR;
+    string = ui_escape(mesg);
+    if (string == NULL)
+        return TCL_ERROR;
 
-	len = strlen(string);
-	scriptlen = sizeof(ui_proc_start) + len + sizeof(ui_proc_end) - 1;
-	script = malloc(scriptlen);
-	if (script == NULL)
-		return TCL_ERROR;
+    len = strlen(string);
+    scriptlen = sizeof(ui_proc_start) + len + sizeof(ui_proc_end) - 1;
+    script = malloc(scriptlen);
+    if (script == NULL)
+        return TCL_ERROR;
 
-	memcpy(script, ui_proc_start, sizeof(ui_proc_start));
-	remaining = scriptlen - sizeof(ui_proc_start);
-	strncat(script, string, remaining);
-	remaining -= len;
-	strncat(script, ui_proc_end, remaining);
-	free(string);
-	rval = Tcl_EvalEx(interp, script, -1, 0);
-	free(script);
-	return rval;
+    memcpy(script, ui_proc_start, sizeof(ui_proc_start));
+    remaining = scriptlen - sizeof(ui_proc_start);
+    strncat(script, string, remaining);
+    remaining -= len;
+    strncat(script, ui_proc_end, remaining);
+    free(string);
+    rval = Tcl_EvalEx(interp, script, -1, 0);
+    free(script);
+    return rval;
 }
 
 int StrsedCmd(ClientData clientData UNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
 {
-	char *pattern, *string, *res;
-	int range[2];
-	Tcl_Obj *tcl_result;
+    char *pattern, *string, *res;
+    int range[2];
+    Tcl_Obj *tcl_result;
 
-	if (objc != 3) {
-		Tcl_WrongNumArgs(interp, 1, objv, "string pattern");
-		return TCL_ERROR;
-	}
+    if (objc != 3) {
+        Tcl_WrongNumArgs(interp, 1, objv, "string pattern");
+        return TCL_ERROR;
+    }
 
-	string = Tcl_GetString(objv[1]);
-	pattern = Tcl_GetString(objv[2]);
-	res = strsed(string, pattern, range);
-	if (!res) {
-		Tcl_SetResult(interp, "strsed failed", TCL_STATIC);
-		return TCL_ERROR;
-	}
-	tcl_result = Tcl_NewStringObj(res, -1);
-	Tcl_SetObjResult(interp, tcl_result);
-	free(res);
-	return TCL_OK;
+    string = Tcl_GetString(objv[1]);
+    pattern = Tcl_GetString(objv[2]);
+    res = strsed(string, pattern, range);
+    if (!res) {
+        Tcl_SetResult(interp, "strsed failed", TCL_STATIC);
+        return TCL_ERROR;
+    }
+    tcl_result = Tcl_NewStringObj(res, -1);
+    Tcl_SetObjResult(interp, tcl_result);
+    free(res);
+    return TCL_OK;
 }
 
 int ExistsuserCmd(ClientData clientData UNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
 {
-	Tcl_Obj *tcl_result;
-	struct passwd *pwent;
-	char *user;
+    Tcl_Obj *tcl_result;
+    struct passwd *pwent;
+    char *user;
 
-	if (objc != 2) {
-		Tcl_WrongNumArgs(interp, 1, objv, "user");
-		return TCL_ERROR;
-	}
+    if (objc != 2) {
+        Tcl_WrongNumArgs(interp, 1, objv, "user");
+        return TCL_ERROR;
+    }
 
-	user = strdup(Tcl_GetString(objv[1]));
-	if (isdigit(*(user)))
-		pwent = getpwuid(strtol(user, 0, 0));
-	else
-		pwent = getpwnam(user);
-	free(user);
+    user = strdup(Tcl_GetString(objv[1]));
+    if (isdigit(*(user)))
+        pwent = getpwuid(strtol(user, 0, 0));
+    else
+        pwent = getpwnam(user);
+    free(user);
 
-	if (pwent == NULL)
-		tcl_result = Tcl_NewIntObj(0);
-	else
-		tcl_result = Tcl_NewIntObj(pwent->pw_uid);
+    if (pwent == NULL)
+        tcl_result = Tcl_NewIntObj(0);
+    else
+        tcl_result = Tcl_NewIntObj(pwent->pw_uid);
 
-	Tcl_SetObjResult(interp, tcl_result);
-	return TCL_OK;
+    Tcl_SetObjResult(interp, tcl_result);
+    return TCL_OK;
 }
 
 int ExistsgroupCmd(ClientData clientData UNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
 {
-	Tcl_Obj *tcl_result;
-	struct group *grent;
-	char *group;
+    Tcl_Obj *tcl_result;
+    struct group *grent;
+    char *group;
 
-	if (objc != 2) {
-		Tcl_WrongNumArgs(interp, 1, objv, "groupname");
-		return TCL_ERROR;
-	}
+    if (objc != 2) {
+        Tcl_WrongNumArgs(interp, 1, objv, "groupname");
+        return TCL_ERROR;
+    }
 
-	group = strdup(Tcl_GetString(objv[1]));
-	if (isdigit(*(group)))
-		grent = getgrgid(strtol(group, 0, 0));
-	else
-		grent = getgrnam(group);
-	free(group);
+    group = strdup(Tcl_GetString(objv[1]));
+    if (isdigit(*(group)))
+        grent = getgrgid(strtol(group, 0, 0));
+    else
+        grent = getgrnam(group);
+    free(group);
 
-	if (grent == NULL)
-		tcl_result = Tcl_NewIntObj(0);
-	else
-		tcl_result = Tcl_NewIntObj(grent->gr_gid);
+    if (grent == NULL)
+        tcl_result = Tcl_NewIntObj(0);
+    else
+        tcl_result = Tcl_NewIntObj(grent->gr_gid);
 
-	Tcl_SetObjResult(interp, tcl_result);
-	return TCL_OK;
+    Tcl_SetObjResult(interp, tcl_result);
+    return TCL_OK;
 }
 
 /* Find the first unused UID > 500
@@ -241,81 +248,81 @@ int ExistsgroupCmd(ClientData clientData UNUSED, Tcl_Interp *interp, int objc, T
    but UIDs < 500 are reserved by Apple */
 int NextuidCmd(ClientData clientData UNUSED, Tcl_Interp *interp, int objc UNUSED, Tcl_Obj *CONST objv[] UNUSED)
 {
-	Tcl_Obj *tcl_result;
-	int cur;
+    Tcl_Obj *tcl_result;
+    int cur;
 
-	cur = MIN_USABLE_UID;
+    cur = MIN_USABLE_UID;
 
-	while (getpwuid(cur) != NULL) {
-		cur++;
-	}
+    while (getpwuid(cur) != NULL) {
+        cur++;
+    }
 
-	tcl_result = Tcl_NewIntObj(cur);
-	Tcl_SetObjResult(interp, tcl_result);
-	return TCL_OK;
+    tcl_result = Tcl_NewIntObj(cur);
+    Tcl_SetObjResult(interp, tcl_result);
+    return TCL_OK;
 }
 
 /* Just as with NextuidCmd, return the first unused gid > 500 */
 int NextgidCmd(ClientData clientData UNUSED, Tcl_Interp *interp, int objc UNUSED, Tcl_Obj *CONST objv[] UNUSED)
 {
-	Tcl_Obj *tcl_result;
-	int cur;
+    Tcl_Obj *tcl_result;
+    int cur;
 
-	cur = MIN_USABLE_GID;
+    cur = MIN_USABLE_GID;
 
-	while (getgrgid(cur) != NULL) {
-		cur++;
-	}
+    while (getgrgid(cur) != NULL) {
+        cur++;
+    }
 
-	tcl_result = Tcl_NewIntObj(cur);
-	Tcl_SetObjResult(interp, tcl_result);
-	return TCL_OK;
+    tcl_result = Tcl_NewIntObj(cur);
+    Tcl_SetObjResult(interp, tcl_result);
+    return TCL_OK;
 }
 
 int UmaskCmd(ClientData clientData UNUSED, Tcl_Interp *interp, int objc UNUSED, Tcl_Obj *CONST objv[] UNUSED)
 {
-	Tcl_Obj *tcl_result;
-	char *tcl_mask, *p;
-	const size_t stringlen = 5; /* 4 digits & \0 */
-	int i;
-	mode_t *set;
-	mode_t newmode;
-	mode_t oldmode;
+    Tcl_Obj *tcl_result;
+    char *tcl_mask, *p;
+    const size_t stringlen = 5; /* 4 digits & \0 */
+    int i;
+    mode_t *set;
+    mode_t newmode;
+    mode_t oldmode;
 
-	if (objc != 2) {
-		Tcl_WrongNumArgs(interp, 1, objv, "mode");
-		return TCL_ERROR;
-	}
+    if (objc != 2) {
+        Tcl_WrongNumArgs(interp, 1, objv, "mode");
+        return TCL_ERROR;
+    }
 
-	tcl_mask = Tcl_GetString(objv[1]);
-	if ((set = setmode(tcl_mask)) == NULL) {
-		Tcl_SetResult(interp, "Invalid umask mode", TCL_STATIC);
-		return TCL_ERROR;
-	}
+    tcl_mask = Tcl_GetString(objv[1]);
+    if ((set = setmode(tcl_mask)) == NULL) {
+        Tcl_SetResult(interp, "Invalid umask mode", TCL_STATIC);
+        return TCL_ERROR;
+    }
 
-	newmode = getmode(set, 0);
-	free(set);
+    newmode = getmode(set, 0);
+    free(set);
 
-	oldmode = umask(newmode);
+    oldmode = umask(newmode);
 
-	tcl_mask = calloc(1, stringlen); /* 4 digits & \0 */
-	if (!tcl_mask) {
-		return TCL_ERROR;
-	}
+    tcl_mask = calloc(1, stringlen); /* 4 digits & \0 */
+    if (!tcl_mask) {
+        return TCL_ERROR;
+    }
 
-	/* Totally gross and cool */
-	p = tcl_mask + stringlen - 1;
-	for (i = stringlen - 1; i > 0; i--) {
-		p--;
-		*p = (oldmode & 7) + '0';
-		oldmode >>= 3;
-	}
+    /* Totally gross and cool */
+    p = tcl_mask + stringlen - 1;
+    for (i = stringlen - 1; i > 0; i--) {
+        p--;
+        *p = (oldmode & 7) + '0';
+        oldmode >>= 3;
+    }
 
-	tcl_result = Tcl_NewStringObj(p, -1);
-	free(tcl_mask);
+    tcl_result = Tcl_NewStringObj(p, -1);
+    free(tcl_mask);
 
-	Tcl_SetObjResult(interp, tcl_result);
-	return TCL_OK;
+    Tcl_SetObjResult(interp, tcl_result);
+    return TCL_OK;
 }
 
 /**
@@ -464,10 +471,107 @@ int lchownCmd(ClientData clientData UNUSED, Tcl_Interp *interp, int objc, Tcl_Ob
     return TCL_OK;
 }
 
+#ifdef __MACH__
+/**
+ * Tcl function to determine whether a file given by path is binary (in terms of being Mach-O)
+ * Defined on Mac-Systems only, because the necessary headers are only available there.
+ *
+ * Synopsis: fileIsBinary filename
+ */
+static int fileIsBinaryCmd(ClientData clientData UNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) {
+    const char *path;
+    FILE *file;
+    uint32_t magic;
+    struct stat st;
+
+    if (objc != 2) {
+        Tcl_WrongNumArgs(interp, 1, objv, "filename");
+        return TCL_ERROR;
+    }
+
+    path = Tcl_GetString(objv[1]);
+    if (-1 == lstat(path, &st)) {
+        /* an error occured */
+        Tcl_SetErrno(errno);
+        Tcl_ResetResult(interp);
+        Tcl_AppendResult(interp, "lstat(", path, "):", (char *)Tcl_PosixError(interp), NULL);
+        return TCL_ERROR;
+    }
+    if (!S_ISREG(st.st_mode)) {
+        /* not a regular file, haven't seen directories which are binaries yet */
+        Tcl_SetObjResult(interp, Tcl_NewBooleanObj(false));
+        return TCL_OK;
+    }
+    if (NULL == (file = fopen(path, "r"))) {
+        Tcl_SetErrno(errno);
+        Tcl_ResetResult(interp);
+        Tcl_AppendResult(interp, "fopen(", path, "): ", (char *)Tcl_PosixError(interp), NULL);
+        return TCL_ERROR;
+    }
+    if (1 != fread(&magic, sizeof(uint32_t), 1, file)) {
+        if (feof(file)) {
+            fclose(file);
+            /* file is shorter than 4 byte, probably not a binary */
+            Tcl_SetObjResult(interp, Tcl_NewBooleanObj(false));
+            return TCL_OK;
+        }
+        /* error while reading */
+        Tcl_SetErrno(errno);
+        Tcl_ResetResult(interp);
+        Tcl_AppendResult(interp, "fread(&magic, 4, 1, ", path, "): ", (char *)Tcl_PosixError(interp), NULL);
+        fclose(file);
+        return TCL_ERROR;
+    }
+    if (magic == MH_MAGIC || magic == MH_MAGIC_64) {
+        fclose(file);
+        /* this is a mach-o file */
+        Tcl_SetObjResult(interp, Tcl_NewBooleanObj(true));
+        return TCL_OK;
+    }
+    if (magic == htonl(FAT_MAGIC)) {
+        uint32_t archcount;
+        /* either universal binary or java class (FAT_MAGIC == 0xcafebabe)
+           see /use/share/file/magic/cafebabe for an explanation of what I'm doing here */
+        if (1 != fread(&archcount, sizeof(uint32_t), 1, file)) {
+            if (feof(file)) {
+                fclose(file);
+                /* file shorter than 8 byte, probably not a binary either */
+                Tcl_SetObjResult(interp, Tcl_NewBooleanObj(false));
+                return TCL_OK;
+            }
+            /* error while reading */
+            Tcl_SetErrno(errno);
+            Tcl_ResetResult(interp);
+            Tcl_AppendResult(interp, "fread(&archcount, 4, 1, ", path, "): ", (char *)Tcl_PosixError(interp), NULL);
+            fclose(file);
+            return TCL_ERROR;
+        }
+
+        /* universal binary header is always big endian */
+        archcount = ntohl(archcount);
+        if (archcount > 0 && archcount < 20) {
+            fclose(file);
+            /* universal binary */
+            Tcl_SetObjResult(interp, Tcl_NewBooleanObj(true));
+            return TCL_OK;
+        }
+
+        fclose(file);
+        /* probably java class */
+        Tcl_SetObjResult(interp, Tcl_NewBooleanObj(false));
+        return TCL_OK;
+    }
+    fclose(file);
+
+    Tcl_SetObjResult(interp, Tcl_NewBooleanObj(false));
+    return TCL_OK;
+}
+#endif
+
 int Pextlib_Init(Tcl_Interp *interp)
 {
-	if (Tcl_InitStubs(interp, "8.4", 0) == NULL)
-		return TCL_ERROR;
+    if (Tcl_InitStubs(interp, "8.4", 0) == NULL)
+        return TCL_ERROR;
 
 	Tcl_CreateObjCommand(interp, "system", SystemCmd, NULL, NULL);
 	Tcl_CreateObjCommand(interp, "flock", FlockCmd, NULL, NULL);
@@ -484,8 +588,11 @@ int Pextlib_Init(Tcl_Interp *interp)
 	Tcl_CreateObjCommand(interp, "xinstall", InstallCmd, NULL, NULL);
 	Tcl_CreateObjCommand(interp, "fs-traverse", FsTraverseCmd, NULL, NULL);
 	Tcl_CreateObjCommand(interp, "filemap", FilemapCmd, NULL, NULL);
-	Tcl_CreateObjCommand(interp, "rpm-vercomp", RPMVercompCmd, NULL, NULL);
-	Tcl_CreateObjCommand(interp, "vercmp", RPMVercompCmd, NULL, NULL);
+#if 1
+	/* the name "rpm-vercomp" is deprecated, use "vercmp" instead */
+	Tcl_CreateObjCommand(interp, "rpm-vercomp", VercompCmd, NULL, NULL);
+#endif
+	Tcl_CreateObjCommand(interp, "vercmp", VercompCmd, NULL, NULL);
 	Tcl_CreateObjCommand(interp, "rmd160", RMD160Cmd, NULL, NULL);
 	Tcl_CreateObjCommand(interp, "sha256", SHA256Cmd, NULL, NULL);
 	Tcl_CreateObjCommand(interp, "base32encode", Base32EncodeCmd, NULL, NULL);
@@ -498,30 +605,33 @@ int Pextlib_Init(Tcl_Interp *interp)
 	Tcl_CreateObjCommand(interp, "unsetenv", UnsetEnvCmd, NULL, NULL);
 	Tcl_CreateObjCommand(interp, "lchown", lchownCmd, NULL, NULL);
 	Tcl_CreateObjCommand(interp, "realpath", RealpathCmd, NULL, NULL);
+#ifdef __MACH__
+    Tcl_CreateObjCommand(interp, "fileIsBinary", fileIsBinaryCmd, NULL, NULL);
+#endif
 
-	Tcl_CreateObjCommand(interp, "readline", ReadlineCmd, NULL, NULL);
-	Tcl_CreateObjCommand(interp, "rl_history", RLHistoryCmd, NULL, NULL);
+    Tcl_CreateObjCommand(interp, "readline", ReadlineCmd, NULL, NULL);
+    Tcl_CreateObjCommand(interp, "rl_history", RLHistoryCmd, NULL, NULL);
 
-	Tcl_CreateObjCommand(interp, "getuid", getuidCmd, NULL, NULL);
-	Tcl_CreateObjCommand(interp, "geteuid", geteuidCmd, NULL, NULL);
-	Tcl_CreateObjCommand(interp, "getgid", getgidCmd, NULL, NULL);
-	Tcl_CreateObjCommand(interp, "getegid", getegidCmd, NULL, NULL);
-	Tcl_CreateObjCommand(interp, "setuid", setuidCmd, NULL, NULL);
-	Tcl_CreateObjCommand(interp, "seteuid", seteuidCmd, NULL, NULL);
-	Tcl_CreateObjCommand(interp, "setgid", setgidCmd, NULL, NULL);
-	Tcl_CreateObjCommand(interp, "setegid", setegidCmd, NULL, NULL);
-	Tcl_CreateObjCommand(interp, "name_to_uid", name_to_uidCmd, NULL, NULL);
-	Tcl_CreateObjCommand(interp, "uid_to_name", uid_to_nameCmd, NULL, NULL);
-	Tcl_CreateObjCommand(interp, "uname_to_gid", uname_to_gidCmd, NULL, NULL);
-	Tcl_CreateObjCommand(interp, "name_to_gid", name_to_gidCmd, NULL, NULL);
-	Tcl_CreateObjCommand(interp, "gid_to_name", gid_to_nameCmd, NULL, NULL);
+    Tcl_CreateObjCommand(interp, "getuid", getuidCmd, NULL, NULL);
+    Tcl_CreateObjCommand(interp, "geteuid", geteuidCmd, NULL, NULL);
+    Tcl_CreateObjCommand(interp, "getgid", getgidCmd, NULL, NULL);
+    Tcl_CreateObjCommand(interp, "getegid", getegidCmd, NULL, NULL);
+    Tcl_CreateObjCommand(interp, "setuid", setuidCmd, NULL, NULL);
+    Tcl_CreateObjCommand(interp, "seteuid", seteuidCmd, NULL, NULL);
+    Tcl_CreateObjCommand(interp, "setgid", setgidCmd, NULL, NULL);
+    Tcl_CreateObjCommand(interp, "setegid", setegidCmd, NULL, NULL);
+    Tcl_CreateObjCommand(interp, "name_to_uid", name_to_uidCmd, NULL, NULL);
+    Tcl_CreateObjCommand(interp, "uid_to_name", uid_to_nameCmd, NULL, NULL);
+    Tcl_CreateObjCommand(interp, "uname_to_gid", uname_to_gidCmd, NULL, NULL);
+    Tcl_CreateObjCommand(interp, "name_to_gid", name_to_gidCmd, NULL, NULL);
+    Tcl_CreateObjCommand(interp, "gid_to_name", gid_to_nameCmd, NULL, NULL);
 
-	Tcl_CreateObjCommand(interp, "tracelib", TracelibCmd, NULL, NULL);
-	Tcl_CreateObjCommand(interp, "isatty", IsattyCmd, NULL, NULL);
-	Tcl_CreateObjCommand(interp, "term_get_size", TermGetSizeCmd, NULL, NULL);
+    Tcl_CreateObjCommand(interp, "tracelib", TracelibCmd, NULL, NULL);
+    Tcl_CreateObjCommand(interp, "isatty", IsattyCmd, NULL, NULL);
+    Tcl_CreateObjCommand(interp, "term_get_size", TermGetSizeCmd, NULL, NULL);
 
-	if (Tcl_PkgProvide(interp, "Pextlib", "1.0") != TCL_OK)
-		return TCL_ERROR;
+    if (Tcl_PkgProvide(interp, "Pextlib", "1.0") != TCL_OK)
+        return TCL_ERROR;
 
-	return TCL_OK;
+    return TCL_OK;
 }
